(* XDS Example Application - Trains *)
(* Copyright (c) 1999 Paul Floyd           *)
<*+ MAIN *>
MODULE Trains;
IMPORT InOut;
TYPE

   (* Base Record Type *)
   (* It also has a "Next" field to allow us to make a linked list *)
   RailwayObject = POINTER TO RailwayObjectDesc;
   RailwayObjectDesc = RECORD
      Next: RailwayObject;
      UnladenWeight: INTEGER;
      MaximumPayload: INTEGER;
      MaximumSpeed: INTEGER;
      Id: INTEGER;
      Description: ARRAY 32 OF CHAR
   END;

   (* Derive a load of other record types from this *)
   TractionUnit = POINTER TO TractionUnitDesc;
   TractionUnitDesc = RECORD (RailwayObjectDesc)
      Power: INTEGER
   END;

   DieselEngine = POINTER TO DieselEngineDesc;
   DieselEngineDesc = RECORD (TractionUnitDesc)
      FuelCapacity: INTEGER
   END;

   ElectricEngine = POINTER TO ElectricEngineDesc;
   ElectricEngineDesc = RECORD (TractionUnitDesc)
      ACMotor, OverheadPickup: BOOLEAN;
   END;

   UnpoweredUnit = POINTER TO UnpoweredUnitDesc;
   UnpoweredUnitDesc = RECORD (RailwayObjectDesc)
   END;

   FreightUnit = POINTER TO FreightUnitDesc;
   FreightUnitDesc = RECORD (UnpoweredUnitDesc)
   END;

   PassengerUnit = POINTER TO PassengerUnitDesc;
   PassengerUnitDesc = RECORD (UnpoweredUnitDesc)
      PassengerCapacity: INTEGER;
   END;

   Tanker = POINTER TO TankerDesc;
   TankerDesc = RECORD (FreightUnitDesc)
      TankContents: ARRAY 32 OF CHAR;
      DangerousSubstance: BOOLEAN
   END;

   Container = POINTER TO ContainerDesc;
   ContainerDesc = RECORD (FreightUnitDesc)
   END;

   CoalWaggon = POINTER TO CoalWaggonDesc;
   CoalWaggonDesc = RECORD (FreightUnitDesc)
   END;

   Carriage = POINTER TO CarriageDesc;
   CarriageDesc = RECORD (PassengerUnitDesc)
      MaximumPassengers: INTEGER;
      FirstClass: BOOLEAN;
   END;

   Sleeper = POINTER TO SleeperDesc;
   SleeperDesc = RECORD (PassengerUnitDesc)
      NumberOfBeds: INTEGER;
   END;

   RestaurantCar = POINTER TO RestaurantCarDesc;
   RestaurantCarDesc = RECORD (PassengerUnitDesc)
      NumberOfTables: INTEGER;
      NumberOfSeats: INTEGER
   END;

VAR
   (* these are the building blocks for building 2 "trains" *)
   ptr: RailwayObject;
   PassengerTrain: ElectricEngine;
   GoodsTrain: DieselEngine;
   PassengerCarriage: Carriage;
   PassengerRestaurantCar: RestaurantCar;
   GoodsTanker: Tanker;
   GoodsCoal: CoalWaggon;
   (* a loop variable *)
   i: INTEGER;

   (* method to return the last item in the linked list *)
   PROCEDURE (ro: RailwayObject)FindLast() : RailwayObject;
   VAR
      TmpPtr: RailwayObject;
   BEGIN
      TmpPtr := ro;
      WHILE TmpPtr.Next # NIL DO
         TmpPtr := TmpPtr.Next
      END; (* endwhile *)
      RETURN TmpPtr
   END FindLast;

   (* method to add an object to the end of the linked list *)
   PROCEDURE (ro: RailwayObject)AddLast(rwo: RailwayObject);
   VAR
      Tmp: RailwayObject;
   BEGIN
      Tmp := ro.FindLast();
      Tmp.Next := rwo;
   END AddLast;

   (* Print information describing this instance *)
   PROCEDURE (ro: RailwayObject)DescribeSelf;
   BEGIN
      InOut.WriteString(ro.Description);
   END DescribeSelf;

   (* Derived method *)
   PROCEDURE (car: Carriage)DescribeSelf;
   BEGIN
      (* call the same method of the immediate base record of this record *)
      car.DescribeSelf^;
      (* write some extra information *)
      IF car.FirstClass = TRUE THEN
         InOut.WriteString(" First Class")
      ELSE
         InOut.WriteString(" Standard Class")
      END; (* endif *)
      InOut.WriteString(" capacity ");
      (* could have imported a module to make this cleaner *)
      InOut.Write(CHR(ORD('0') + car.MaximumPassengers DIV 10));
      InOut.Write(CHR(ORD('0') + car.MaximumPassengers MOD 10));
      InOut.WriteString(" people")
   END DescribeSelf;

   PROCEDURE (grub: RestaurantCar)DescribeSelf;
   BEGIN
      grub.DescribeSelf^;
      InOut.WriteString(" tables: ");
      InOut.Write(CHR(ORD('0') + grub.NumberOfTables DIV 10));
      InOut.Write(CHR(ORD('0') + grub.NumberOfTables MOD 10));
      InOut.WriteString(" seats: ");
      InOut.Write(CHR(ORD('0') + grub.NumberOfSeats DIV 10));
      InOut.Write(CHR(ORD('0') + grub.NumberOfSeats MOD 10))
   END DescribeSelf;

BEGIN
   (* main *)

   (* create a goods train *)
   (* don't do anything with this though *)
   NEW(GoodsTrain);

   FOR i := 0 TO 9 DO
      NEW(GoodsTanker);
      GoodsTrain.AddLast(GoodsTanker)
   END; (* endfor *)

   FOR i := 0 TO 4 DO
      NEW(GoodsCoal);
      GoodsTrain.AddLast(GoodsCoal)
   END; (* endfor *)

   (* create a passenger train *)
   NEW(PassengerTrain);
   COPY("Leccy Motor", PassengerTrain.Description);

   (* add 15 standard class carriages *)
   FOR i := 0 TO 14 DO
      NEW(PassengerCarriage);
      PassengerCarriage.FirstClass := FALSE;
      COPY("Plebs Carriage ", PassengerCarriage.Description);
      PassengerCarriage.Description[15] := CHR(ORD('A') + i);
      PassengerCarriage.MaximumPassengers := 80;
      PassengerTrain.AddLast(PassengerCarriage)
   END; (* endfor *)

   (* add a buffet car *)
   NEW(PassengerRestaurantCar);
   COPY("Beer Waggon", PassengerRestaurantCar.Description);
   PassengerRestaurantCar.NumberOfTables := 10;
   PassengerRestaurantCar.NumberOfSeats := 24; (* 8*2 + 2*4 *)
   PassengerTrain.AddLast(PassengerRestaurantCar);

   (* and a single first class carriage *)
   NEW(PassengerCarriage);
   PassengerCarriage.FirstClass := TRUE;
   COPY("Posh Carriage", PassengerCarriage.Description);
   PassengerCarriage.MaximumPassengers := 60;
   PassengerTrain.AddLast(PassengerCarriage);

   (* a temporary pointer for traversing the linked list *)
   ptr := PassengerTrain;

   (* visit each node and print information about it *)
   REPEAT
      ptr.DescribeSelf;
      InOut.WriteLn;
      ptr := ptr.Next
   UNTIL ptr = NIL
END Trains.
